import time

from licensing.methods import Helpers, Key
from sphinx.util import logging

from sphinx_needs_enterprise.config import (
    API_TOKEN,
    LICENSE_INTERVAL_SECS,
    LICENSE_RETRIES,
    LICENSE_RETRY_SECS,
    RSA_PUB_KEY,
    TEXT_INVALID,
    TEXT_INVALID_WARNING,
    TEXT_PRIVATE_FULL,
    TEXT_PRIVATE_SHORT,
    TEXT_VALID,
)


class License:
    def __init__(
        self,
        product_id,
        product_name,
        license_key=None,
        product_url=None,
        docs_url=None,
        license_url=None,
        suppress_private_message=False,
        rsa_key=RSA_PUB_KEY,
        api_token=API_TOKEN,
    ):

        self.log = logging.getLogger(__name__)

        self.rsa_pub_key = rsa_key
        self.token = api_token
        self.product_id = product_id
        self.product_name = product_name
        self.product_url = product_url
        self.docs_url = docs_url
        self.license_url = license_url
        self.license_key = license_key
        self.suppress_private_message = suppress_private_message

        self.machine_code = Helpers.GetMachineCode()

        self.is_valid = None
        self.activated = False
        self.is_private = None
        self.is_commercial = None
        self.customer = None
        self.server_reachable = None
        self.license = None
        self.message = None
        self.limit_reached = None
        self.last_check = 0

    def check(self, own_output=False):
        """
        Checks and activates a given key.
        This function should be used as early as possible.

        Set "own_output" to True to let the extension print all the needed output to inform the user about the
        license status. Otherwise some common information will be printed, if license check fails.

        :param own_output: Flag, if True no output will be printed.
        :return: True, if license is valid and available, else false.
        """

        # check/recheck a license only, if the license-timeout as been reached.
        # Otherwise a check makes no sense and is too early.
        if not time.time() - self.last_check >= LICENSE_INTERVAL_SECS:
            return self.is_valid, self.server_reachable

        self.last_check = time.time()

        self.is_valid = False
        self.activated = False
        self.server_reachable = None

        if not self.license_key or self.license_key.upper() == "PRIVATE":
            self.is_valid = True
            self.is_private = True
            self.is_commercial = False
            self.server_reachable = True
            self.limit_reached = False

            return self.is_valid, self.server_reachable

        self.is_private = False
        self.is_commercial = True

        attempt = 1
        while (self.limit_reached is None or self.limit_reached is True) and attempt <= 3:
            if attempt > 1:
                time.sleep(LICENSE_RETRY_SECS)
            self.log.debug(f"License attempt: {attempt}")
            self.license, self.message = Key.activate(
                token=self.token,
                rsa_pub_key=self.rsa_pub_key,
                product_id=self.product_id,
                key=self.license_key,
                machine_code=self.machine_code,
                floating_time_interval=LICENSE_INTERVAL_SECS,
            )

            if "Could not contact the server" in self.message:
                self.server_reachable = False
                self.log.info(
                    f"License server not reachable "
                    f"Try again in {LICENSE_RETRY_SECS} secs. Attempt {attempt}/{LICENSE_RETRIES}"
                )
            else:
                self.server_reachable = True

            if "limit has been reached" in self.message:
                self.limit_reached = True
                self.log.info(
                    f"License limit has been reached. "
                    f"Try again in {LICENSE_RETRY_SECS} secs. Attempt {attempt}/{LICENSE_RETRIES}"
                )
            else:
                self.limit_reached = False

            attempt += 1

        if self.license is None:
            return self.is_valid, self.server_reachable

        self.is_valid = True
        self.activated = True

        self.customer = f"{self.license.customer['Name']} ({self.license.customer['CompanyName']})"

        return self.is_valid, self.server_reachable

    def free(self):
        """
        Gives back a a floating license.
        Should normally be called at the end of a build process
        """

        # Private licenses are not registered on server.
        # Also do not free a license, if it is invalid.
        if self.is_private or not self.is_valid:
            return

        success, message = Key.deactivate(
            token=self.token,
            product_id=self.product_id,
            key=self.license_key,
            machine_code=self.machine_code,
            floating=True,
        )
        if not success:
            self.log.info(f"Could not free license: {message}")

        return success

    def print_info(self):
        """
        Prints the current user information regarding the license status.

        As the status is checked several times during a single build,
        we can not print everytime information to the user.
        So this functions helps to bother the user only on specific points of the build process
        (e.g. start and end).
        """
        if not self.license_key or self.is_private:
            if not self.suppress_private_message:
                self.log.info(
                    TEXT_PRIVATE_FULL.format(
                        product_name=self.product_name,
                        product_url=self.product_url,
                        docs_url=self.docs_url,
                        license_url=self.license_url,
                    )
                )
            else:
                self.log.info(TEXT_PRIVATE_SHORT.format(product_name=self.product_name))
            return

        if self.license is None:
            if not self.server_reachable:
                self.log.info(f"License server not reachable to validate license for {self.product_name}.")
            else:

                self.log.warning(
                    TEXT_INVALID_WARNING.format(
                        license_key=self.license_key, product_name=self.product_name, message=self.message
                    )
                )
                self.log.info(
                    TEXT_INVALID.format(
                        license_key=self.license_key,
                        product_name=self.product_name,
                        message=self.message,
                        product_url=self.product_url,
                        docs_url=self.docs_url,
                    )
                )
        else:
            self.log.info(TEXT_VALID.format(product_name=self.product_name, customer=self.customer))



import json
import os
import subprocess
import webbrowser
from datetime import datetime, timedelta

import click
import elasticsearch
import jinja2
from tqdm import tqdm

# API has changed with Sphinx-Needs version 0.7.3
try:
    from sphinxcontrib.needs.utils import NeedsList
except ImportError:
    from sphinxcontrib.needs.needsfile import NeedsList

from sphinx_needs_enterprise.scripts.loader import service_loader


@click.group()
def cli():
    pass


@cli.command(name="import")
@click.argument("service")
@click.option("-c", "--conf", default="conf.py", type=click.Path(exists=True), help="Relative path to conf.py")
@click.option("-o", "--outdir", default=".", type=str, help="Relative path to folder for needs.json")
@click.option("-q", "--query", type=str, help="Query string")
@click.option(
    "-r", "--reuse", "old_needfile", type=click.Path(exists=True), help="Relative path to an existing needs.json file"
)
@click.option("-v", "--version", type=str, help="Version under which data shall be stored")
@click.option("-w", "--wipe", is_flag=True, default=False, help="Erases all data from reused file for used version.")
def import_cmd(service, conf, outdir, query, old_needfile, version, wipe):
    conf_path = os.path.abspath(conf)

    service_obj, sphinx_config = service_loader(service, conf_path)
    options = {}
    if query:
        options["query"] = query
    params = service_obj._prepare_request(options)

    # Getting data
    click.echo()
    click.echo(f'URL: {params["url"]}')
    click.echo(f'Query: {params["query"]}')
    click.echo("Sending request:  ", nl=False)
    data = service_obj.request(options)
    click.echo("Done")
    click.echo(f"Retrieved {len(data)} elements")

    # Storing data
    os.makedirs(outdir, exist_ok=True)
    needlist = NeedsList(sphinx_config, outdir, ".")
    version = version or needlist.current_version
    click.echo(f"Version to use: {version}")

    if old_needfile:
        if not os.path.isabs(old_needfile):
            old_needfile = os.path.abspath(old_needfile)
        click.echo(f"Reusing needs.json: {old_needfile}")
        needlist.load_json(old_needfile)

    if wipe:
        click.echo(f"Erasing existing data for version {version}.")
        needlist.wipe_version(version)

    for datum in data:
        needlist.add_need(version, datum)

    click.echo("\nStoring data to json file: ", nl=False)
    needlist.write_json()
    click.echo("Done")


@cli.command(name="export")
@click.argument("service")
@click.option(
    "-j", "--json", "needs_file", default="needs.json", type=click.Path(), help="Relative path to a needs.json file"
)
@click.option("-c", "--conf", default="conf.py", type=click.Path(exists=True), help="Relative path to conf.py")
@click.option("-v", "--version", type=str, help="Version under which data shall be stored")
@click.option("-x", "--extra", multiple=True, type=click.Tuple([str, str]))
@click.option("-h", "--hours", type=int, default="0")
@click.option("-s", "--skip", type=int, default="0")
def export_cmd(service, needs_file, conf, version, extra, hours, skip):
    conf_path = os.path.abspath(conf)

    service_obj, sphinx_config = service_loader(service, conf_path)

    # Load needs.json data
    click.echo("\nReading json data: ", nl=False)
    with open(needs_file, "rb") as json_file:
        needs_data = json.load(json_file)
    click.echo("Done")

    if not version:
        version = needs_data["current_version"]
    needs = needs_data["versions"][version]["needs"]

    url = service_obj.url
    click.echo(f"Connectiong to Elasticsearch url: {url}")
    es = elasticsearch.Elasticsearch([url])

    es_index = service_obj.index
    if not es.indices.exists(index=es_index):
        click.echo(f"Creating index {es_index}")
        es.indices.create(es_index)
    else:
        click.echo(f"Using index {es_index}")

    elements = tqdm(needs.items(), bar_format="{bar:80} {percentage:3.0f}% | {desc}")
    counter = 0
    for element_index, element in enumerate(elements):
        if skip > 0:
            if element_index % skip == 0:
                continue
        key, need = element
        # click.echo(f'{key}')
        # need.pop("id", None)  # Remove id, as it is used by ElasticSearch internally
        for cus in extra:
            need[cus[0]] = cus[1]
        need["uploaded_at"] = datetime.now() + timedelta(hours=hours)
        elements.set_description(f"Uploading need {key}")
        es.index(index=es_index, document=need)
        counter += 1

    click.echo(f"Uploaded {counter} elements.")

    es.indices.refresh(index=es_index)


@cli.command()
@click.option(
    "-j", "--json", "needs_file", default="needs.json", type=click.Path(), help="Relative path to a needs.json file"
)
@click.option(
    "-t", "--template", "template_path", type=click.Path(exists=True), help="Relative path to a template file"
)
@click.option(
    "-o", "--output", "output_path", default="needs.rst", type=click.Path(), help="Relative path to output file"
)
def render(needs_file, template_path, output_path):
    if template_path is None:
        template_path = os.path.join(os.path.dirname(__file__), "../", "templates", "needs.rst.template")

    needs_file = os.path.abspath(needs_file)
    template_path = os.path.abspath(template_path)
    template_file = os.path.basename(template_path)
    template_dir = os.path.dirname(template_path)
    output_path = os.path.abspath(output_path)
    output_dir = os.path.dirname(output_path)

    click.echo(f"Needs file: {needs_file}")
    click.echo(f"Template file: {template_path}")
    click.echo(f"Output file: {output_path}\n")

    if not os.path.exists(needs_file):
        click.echo(f"Error: Needs file does not exist: {needs_file}")

    if not os.path.exists(template_path):
        click.echo(f"Error: Template file does not exist: {template_path}")

    # Load needs.json data
    click.echo("\nReading json data: ", nl=False)
    with open(needs_file, "rb") as json_file:
        need_data = json.load(json_file)
    click.echo("Done")

    # Render template
    click.echo("Rendering template: ", nl=False)
    template_loader = jinja2.FileSystemLoader(searchpath=template_dir)
    template_env = jinja2.Environment(loader=template_loader)
    template = template_env.get_template(f"{template_file}".strip())

    output_data = template.render(data=need_data, now=datetime.now())

    click.echo("Done")

    # Storing data
    click.echo("Storing data: ", nl=False)
    os.makedirs(output_dir, exist_ok=True)
    with open(output_path, "w") as output_fp:
        output_fp.write(output_data)

    click.echo("Done")

    click.echo("All good, we are done! 🎉")


@cli.group()
def dev():
    pass


@dev.command(name="docker")
@click.argument("operation")
@click.option("-b", "--browser", is_flag=True, default=False, help="Opens a browser for each service")
def docker(operation, browser):
    """
    Starts all docker configurations delivered by Sphinx-Needs Enterprise for development.

    These are mainly containers for JIRA, CodeBeamer and more to test the different functions of
    Sphinx-Needs Enterprise.

    The OPERATION argument must be one of: start, up, stop, down.
    """
    if operation.lower() not in ["start", "up", "stop", "down"]:
        click.echo("operation must be one of: start, up, stop, down")
        return 1

    operation = operation.lower()

    click.echo("Starting all docker configurations")

    docker_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "../", "../", "docker/"))

    # Collect docker-compose files
    docker_files = []
    for element in os.listdir(docker_path):
        subelement = os.path.join(docker_path, element)
        if os.path.isdir(subelement):
            subfolder_file = os.path.join(subelement, "docker-compose.yml")
            if os.path.exists(subfolder_file):
                docker_files.append(subfolder_file)

    click.echo("Found following docker configurations:")

    args = ["docker-compose"]
    for docker_file in docker_files:
        args.append("-f")
        args.append(f"{docker_file}")
        click.echo(f" {docker_file}")
    args.append(operation)
    if operation in ["up", "start"]:
        args.append("-d")

    click.echo(f'\nExecuting {" ".join(args)}')
    os.chdir(docker_path)
    click.echo(f"CWD: {os.getcwd()}")

    subprocess.run(args)

    if browser:
        click.echo("Opening web browsers")
        for index, docker_file in enumerate(docker_files):
            url = f"127.0.0.1:{8080 + index}"
            click.echo(f"Opening {url} for {docker_file}")
            webbrowser.open(url)

    click.echo("All good, we are done! 🎉")


if "main" in __name__:
    cli()



import logging
import operator
from functools import reduce  # forward compatibility for Python 3

log = logging.getLogger(__name__)


def dict_get(root, items, default=None):
    """
    Access a nested object in root by item sequence.

    Usage::
       data = {"nested": {"a_list": [{"finally": "target_data"}]}}
       value = dict_get(["nested", "a_list", 0, "finally"], "Not_found")

    """
    try:
        value = reduce(operator.getitem, items, root)
    except (KeyError, IndexError, TypeError) as e:
        log.debug(e)
        return default
    return value


def dict_undefined_set(dict_obj, key, value):
    if key not in dict_obj:
        dict_obj[key] = value



import re

import requests
from jinja2 import Template
from sphinxcontrib.needs.services.base import BaseService

from sphinx_needs_enterprise.exceptions import (
    CommunicationException,
    InvalidConfigException,
    LicenseException,
)
from sphinx_needs_enterprise.license import License
from sphinx_needs_enterprise.util import dict_get


class ServiceExtension(BaseService):
    def __init__(
        self,
        config,
        url=None,
        url_postfix=None,
        query="",
        query_postfix="",
        user="",
        password="",
        token="",
        id_prefix="",
        content=None,
        mappings=None,
        mappings_replaces=None,
        extra_data=None,
        **kwargs,
    ):

        self.config = config

        self.url = url or config.get("url", "")
        self.url_postfix = url_postfix or config.get("url_postfix", "")

        self.query = query or config.get("query", "")
        self.query_postfix = query_postfix or config.get("query_postfix", "")

        self.user = user or config.get("user", "")
        self.password = password or config.get("password", "")
        self.token = token or config.get("token", "")

        self.id_prefix = id_prefix or config.get("id_prefix", "")

        self.content = content or config.get("content", "")

        self.mappings = mappings or config.get("mappings", {})
        self.mapping_replaces = mappings_replaces or config.get("mappings_replaces", {})
        self.extra_data = extra_data or config.get("extra_data", {})

        self.license_key = None
        self.product_id = None
        self.product_name = None
        self.license = None

        super().__init__(**kwargs)

    def set_license(
        self,
        product_id,
        product_name,
        product_url,
        docs_url,
        license_key=None,
        private_feature="f1",
        commercial_feature="f2",
        suppress_private_message=False,
    ):
        key = license_key or self.config.get("license_key", None)
        if key is None:
            raise LicenseException("No license key provided.")

        self.license_key = key
        self.product_id = product_id
        self.product_name = product_name
        self.product_url = product_url
        self.docs_url = docs_url
        self.suppress_private_message = suppress_private_message

        self.license = License(
            self.product_id,
            self.product_name,
            self.license_key,
            product_url=product_url,
            docs_url=docs_url,
            suppress_private_message=suppress_private_message,
        )

    def check_license(self):
        if self.license is None:
            raise LicenseException(
                'License not defined. User "set_license" to activate license ' "support for the Sphinx-Needs extension."
            )

        self.license.check()

    def _prepare_request(self, options):
        if options is None:
            options = {}
        url = options.get("url", self.url)
        url = url + self.url_postfix

        auth = (options.get("user", self.user), options.get("password", self.password))
        query = options.get("query", self.query)
        query = query + self.query_postfix

        request = {"url": url, "auth": auth, "query": query, "params": {}}
        return request

    def _send_request(self, request):
        """
        Sends the final request.

        ``request`` must be a dictionary, which contains data valid to the requests-lib.
        This request-data gets mostly defined by using ``prepare_request``.

        :param request: dict
        :return: request answer
        """

        # params = {
        #     'queryString': query
        # }

        result = requests.request(**request)
        if result.status_code >= 300:
            raise CommunicationException(f"Problems accessing {result.url}.\nReason: {result.text}")

        return result

    def _extract_data(self, data, options):
        """
        Extract data of a list/dictionary, which was retrieved via send_request.

        :param data: list or dict
        :param options: dict of set directive options
        :return: list of need-data
        """

        need_data = []
        if options is None:
            options = {}
        for item in data:
            extra_data = {}
            for name, selector in self.extra_data.items():
                if not (isinstance(selector, tuple) or isinstance(selector, list) or isinstance(selector, str)):
                    raise InvalidConfigException(
                        f"Given selector for {name} of extra_data must be a list or tuple. "
                        f'Got {type(selector)} with value "{selector}"'
                    )
                if isinstance(selector, str):
                    # Set the "hard-coded" string as value
                    extra_data[name] = selector
                else:
                    extra_data[name] = dict_get(item, selector)

            for regex, new_str in self.mapping_replaces.items():
                self.content = re.sub(regex, new_str, self.content)
            content_template = Template(self.content)
            context = {"data": item, "options": options}
            content = content_template.render(context)
            content += "\n\n| \n"  # Add enough space between content and extra_data

            # Add extra_data to content
            for key, value in extra_data.items():
                content += f"\n| **{key}**: {value}"
            content += "\n"

            prefix = options.get("prefix", self.id_prefix)

            need_values = {}
            for name, selector in self.mappings.items():
                if not (isinstance(selector, tuple) or isinstance(selector, list) or isinstance(selector, str)):
                    raise InvalidConfigException(
                        f"Given selector for {name} of mapping must be a list or tuple. "
                        f'Got {type(selector)} with value "{selector}"'
                    )
                if isinstance(selector, str):
                    # Set the "hard-coded" string as value
                    need_values[name] = selector
                else:
                    need_values[name] = str(dict_get(item, selector))

                for regex, new_str in self.mapping_replaces.items():
                    need_values[name] = re.sub(regex, new_str, need_values[name])

                if name == "id":
                    need_values[name] = prefix + need_values[name]

            finale_data = {"content": content}
            finale_data.update(need_values)

            need_data.append(finale_data)
        return need_data



import json

from sphinx_needs_enterprise.extensions.extension import ServiceExtension
from sphinx_needs_enterprise.util import dict_undefined_set

DEFAULT_CONTENT = """
{% set desc_list = data.description.split('\n') %}
{% if options.raw == "True"%}
.. code-block:: text
   {% for line in desc_list %}
   {{line}}
   {%- endfor %}
{% elif options.wiki2html == "True" %}
.. raw:: html
   {% for line in desc_list %}
   {{line}}
   {%- endfor %}

{% else %}
{{data.description}}
{% endif %}
"""


class CodebeamerService(ServiceExtension):
    options = ["query", "prefix", "raw", "wiki2html"]

    def __init__(self, app, name, config, **kwargs):
        self.app = app
        self.name = name

        # Set default values, if nothing got configured
        dict_undefined_set(config, "url", "http://127.0.0.1:8080")
        dict_undefined_set(config, "query", "")
        dict_undefined_set(config, "id_prefix", "CB_")
        dict_undefined_set(config, "url_postfix", "/rest/v3/items/query")
        dict_undefined_set(config, "content", DEFAULT_CONTENT)
        dict_undefined_set(config, "raw", "False")
        dict_undefined_set(config, "wiki2html", "True")

        mappings_default = {
            "id": ["id"],
            "type": ["typeName"],
            "status": ["status", "name"],
            "title": ["name"],
        }
        dict_undefined_set(config, "mappings", mappings_default)

        mappings_replaces_default = {
            r"^Task$": "task",
            r"^Requirement$": "req",
            r"^Specification$": "spec",
            r"\\\\\\\\\\r\\n": "\n\n",
        }
        dict_undefined_set(config, "mappings_replaces", mappings_replaces_default)

        super().__init__(config, **kwargs)

    def request(self, options=None):
        wiki2html = options.get("wiki2html", str(self.config["wiki2html"]))
        options["wiki2html"] = wiki2html  # Just to be sure that there is a value

        raw = options.get("raw", str(self.config["raw"]))
        options["raw"] = raw  # Just to be sure that there is a value

        params = self._prepare_request(options)

        request_params = {
            "method": "GET",
            "url": params["url"],
            "auth": params["auth"],
            "params": {
                "queryString": params["query"],
                "descriptionFormat": "HTML",
                "descFormat": "HTML",
            },
        }

        answer = self._send_request(request_params)
        data = answer.json()["items"]
        for datum in data:
            # Be sure "description" is set and valid
            if "description" not in datum or datum["description"] is None:
                datum["description"] = ""
            elif datum["descriptionFormat"] == "Wiki" and wiki2html == "True":
                # Transform the Codebeamer wiki syntax to HTML.
                # Must be done by an API request for each item.
                url = options.get("url", self.url)
                url = url + "/api/v3/projects/2/wiki2html"

                auth = (options.get("user", self.user), options.get("password", self.password))

                wiki2html_params = {
                    "method": "POST",
                    "url": url,
                    "auth": auth,
                    "headers": {"Content-Type": "application/json", "accept": "text/html"},
                    "data": json.dumps({"markup": datum["description"]}),
                }

                wiki2html_answer = self._send_request(wiki2html_params)
                datum["description"] = wiki2html_answer.text

        need_data = self._extract_data(data, options)

        return need_data


class InvalidConfigException(BaseException):
    pass



def get_app(sphinx_config):
    """
    Create a dummy app object, which looks like a Sphinx app and contains all needed
    information for the Services classes.

    This helps us to reuse the code from services/jira and co.
    """

    class App:
        def __init__(self):
            self.config = {
                "needs_services": sphinx_config.needs_services,
                "needs_types": getattr(sphinx_config, "needs_types", [{"directive": "req"}]),
                "needs_is_private_project": getattr(sphinx_config, "needs_is_private_project", False),
            }

    return App()



import importlib
import os
import re
import sys
import types

import click

from sphinx_needs_enterprise.config import get_providers
from sphinx_needs_enterprise.scripts.utils import get_app


def service_loader(service, conf_path):
    """
    Loads the service form a given service_name in a given conf.py file.
    """
    given_service = service
    conf_path = os.path.abspath(conf_path)

    click.echo(f"Importing config from {conf_path}")
    # Import the conf.py file as config
    loader = importlib.machinery.SourceFileLoader("conf", conf_path)
    sphinx_config = types.ModuleType(loader.name)
    loader.exec_module(sphinx_config)

    if not hasattr(sphinx_config, "needs_services"):
        click.echo("Missing needs_services in configuration")
        sys.exit(1)

    configured_services = sphinx_config.needs_services
    if given_service not in configured_services:
        click.echo(f'Unknown service: {given_service}. Available services are: {", ".join(configured_services)}')
        sys.exit(1)

    config = configured_services[given_service]

    service = None
    for name, provider in get_providers().items():
        if re.search(provider["regex"], given_service):
            service = provider["service"]
            click.echo(f'Using provider "{name}" for given service {given_service}')
            break

    if not service:
        click.echo(f"Could not find a matching service provider for {given_service}. Known providers are:")
        for name, provider in get_providers().items():
            click.echo(f'  {name} with regex {provider["regex"]}')
        sys.exit(1)

    app = get_app(sphinx_config)
    service_obj = service(app, given_service, config)

    return service_obj, sphinx_config



def get_providers():
    from sphinx_needs_enterprise.services.azure import AzureService
    from sphinx_needs_enterprise.services.codebeamer import CodebeamerService
    from sphinx_needs_enterprise.services.elasticsearch import ElasticsearchService
    from sphinx_needs_enterprise.services.jira import JiraService

    providers = {
        "jira": {"service": JiraService, "regex": "^jira"},
        "azure": {"service": AzureService, "regex": "^azure"},
        "codebeamer": {"service": CodebeamerService, "regex": "^codebeamer|^cb"},
        "elasticsearch": {"service": ElasticsearchService, "regex": "^elastic|^es"},
    }
    return providers


# LICENSE
RSA_PUB_KEY = "<RSAKeyValue><Modulus>t5w/Rj4SijVJyQALRUzEJmV8Vin4P0SRCAqWSnttULrcIvkiEMhVS7OnUwUqWeRvrpDyvRy53hGHTgPIQG8pUrh0ZpX3KLRmEqBvtNEU6hOjALowi0Q/zEqYmahCh2fTb8k9FAfcMmO70TMIFfvYYGmogsmdrrCct+xEbGDM01Hhu+Go/61Jb8PpyRmZOgza5fbBC4v4rr9WF5wK6Jui7+P31bNSZDkDuxmx7dd5G+Gk2KImkRPxTZMBVQ6WmFiWFsac/KNvRskoKGlKQTaPUhpxsiPQgLnUVJPy4tzNL1qBzCXCMW3L2dgBM6W8j1kv6WbbouNbke6N+U3DIVehyQ==</Modulus><Exponent>AQAB</Exponent></RSAKeyValue>"
API_TOKEN = "WyI0MDg4MjQyIiwibmJ4Y2FWeEtmRjIrMG8xZ2szeis2NGlZU09meDk5RGFGdG1vNGJjNSJd"

SNE_PRODUCT_ID = 13000
SNE_PRODUCT_NAME = "Sphinx-Needs Enterprise"
SNE_PRODUCT_URL = "https://sphinx-needs.com"
SNE_DOCS_URL = "https://useblocks.com/sphinx-needs-enterprise"
SNE_LICENSE_URL = "http://useblocks.com/sphinx-needs-enterprise/license.html"

LICENSE_RETRIES = 3
LICENSE_RETRY_SECS = 5
LICENSE_INTERVAL_SECS = 300


# TEXT MESSAGES
TEXT_PRIVATE_FULL = """
* Sphinx-Needs Enterprise License Information ************************************************
* No commercial license configured!
* This allows the free usage of {product_name} for private Sphinx projects
* of any size.
* If this is a academic or commercial project, please obtain a license under
* {product_url}.
* You can hide this message by setting "needs_enterprise_license = "PRIVATE"
* or by setting a valid commercial license.
* For technical support visit {docs_url}.
* For license information visit {license_url}.
***********************************************************************************************\n"""

TEXT_PRIVATE_SHORT = "Using {product_name} for private projects."

TEXT_INVALID = """
* Sphinx-Needs Enterprise License Information ************************************************
* Provided license '{license_key}' for extension {product_name} not valid.
* Reason: {message}
* Remove the license from configuration to activate the private usage.
* To obtain a valid license: {product_url}.
* For technical details please visit {docs_url}
***********************************************************************************************
"""
TEXT_INVALID_WARNING = """
Provided license "{license_key}" for extension {product_name} not valid.
Reason: {message}
"""

TEXT_VALID = "Commercial license for {product_name} by {customer} validated."



from sphinx_needs_enterprise.extensions.extension import ServiceExtension
from sphinx_needs_enterprise.util import dict_undefined_set

DEFAULT_FIELDS = [
    "System.Id",
    "System.WorkItemType",
    "System.Title",
    "System.State",
    "System.AreaPath",
    "System.IterationPath",
    "System.Tags",
]

DEFAULT_ORDER = "[System.ChangedDate] desc"


class ElasticsearchService(ServiceExtension):
    options = ["query", "fields", "prefix"]

    def __init__(self, app, name, config, **kwargs):
        self.app = app
        self.name = name

        # Set default values, if nothing got configured
        dict_undefined_set(config, "url", "https://dev.azure.com/YOURORG")
        dict_undefined_set(config, "query", "[System.WorkItemType] = 'Issue'")
        dict_undefined_set(config, "fields", DEFAULT_FIELDS)
        dict_undefined_set(config, "order", DEFAULT_ORDER)
        dict_undefined_set(config, "id_prefix", "AZURE_")
        dict_undefined_set(config, "content", "{{fields.description}}")
        dict_undefined_set(config, "index", "needs")

        # If no mapping is given, we need to use a need-type, which really exists.
        # So we take the first one from config, instead of hard-coding a value or
        # getting it from the jira issue.
        default_need_type = str(self.app.config["needs_types"][0]["directive"])

        self.index = config["index"]

        mappings_default = {
            "id": ["key"],
            "type": default_need_type,
            "title": ["fields", "summary"],
            "status": ["fields", "status", "name"],
        }
        dict_undefined_set(config, "mappings", mappings_default)

        mappings_replaces_default = {}
        dict_undefined_set(config, "mappings_replaces", mappings_replaces_default)

        super().__init__(config, **kwargs)

    def request(self, options=None):
        """
        Request data from Azure devops.

        This is not calling the REST API on its own, instead we use the azure python client.
        See: https://github.com/microsoft/azure-devops-python-api
        """

        query = options.get("query", self.config["query"])
        fields = options.get("fields", self.config["fields"])
        order = options.get("order", self.config["order"])

        wiql = self._create_wiql(query, fields, order)

        wiql_results = self.client.query_by_wiql(wiql, top=200).work_items

        cleaned_data = []
        if wiql_results:
            # The wiql_result contain references only, so we need to get the data for each item step by step... *sigh*
            for res in wiql_results:
                datum = self.client.get_work_item(int(res.id))

                # The returned datum object returns functions and other stuff, we don't need.
                # See we store only elements, which we need (but try to keep the original structure)
                cleaned_datum = {
                    "id": datum.id,
                    "rev": datum.rev,
                    "url": datum.url,
                    "fields": datum.fields,
                    "relations": datum.relations,
                }
                cleaned_data.append(cleaned_datum)

        need_data = self._extract_data(cleaned_data, options)

        return need_data


class InvalidConfigException(BaseException):
    pass



import re

from sphinx_needs_enterprise.config import (
    SNE_DOCS_URL,
    SNE_LICENSE_URL,
    SNE_PRODUCT_ID,
    SNE_PRODUCT_NAME,
    SNE_PRODUCT_URL,
    get_providers,
)
from sphinx_needs_enterprise.license import License
from sphinx_needs_enterprise.version import VERSION


def setup(app):
    app.add_config_value("needs_enterprise_license", "", "html", types=[str])
    # Register sphinx-needs stuff after it has been initialised.
    app.connect("env-before-read-docs", prepare_env)
    app.connect("source-read", process_per_doc)
    app.connect("doctree-resolved", process_per_doc)
    app.connect("build-finished", process_finish)

    return {
        "version": VERSION,  # identifies the version of our extension
        "parallel_read_safe": True,  # support parallel modes
        "parallel_write_safe": True,
    }


def prepare_env(app, env, _docname):
    license_key = getattr(app.config, "needs_enterprise_license", "")
    if license_key.upper() == "PRIVATE":
        suppress_private_message = True
    else:
        suppress_private_message = False

    sne_license = License(
        SNE_PRODUCT_ID,
        SNE_PRODUCT_NAME,
        license_key,
        SNE_PRODUCT_URL,
        SNE_DOCS_URL,
        SNE_LICENSE_URL,
        suppress_private_message,
    )

    env.needs_sne_license = sne_license

    env.needs_sne_license.check()
    env.needs_sne_license.print_info()

    for service in getattr(app.config, "needs_services", {}).keys():
        for name, provider in get_providers().items():
            if re.search(provider["regex"], service):
                service_provider_class = provider["service"]
                app.needs_services.register(service, service_provider_class)
                break  # check next configured service


def process_per_doc(app, *kwargs):
    """
    Check the license everytime a new rst-file is read or get written.
    The license internal logic cares about not requesting information
    from license server too often.
    """
    sne_license = app.env.needs_sne_license
    sne_license.check()


def process_finish(app, exception):
    """
    Finally check/print again license after everything was done
    """
    # There are use cases, where needs_sne_license was not set up. (e.g. no files to build or prior error)
    if hasattr(app.env, "needs_sne_license"):
        app.env.needs_sne_license.print_info()
        app.env.needs_sne_license.free()  # Give back license



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sphinx_needs_enterprise',
 'sphinx_needs_enterprise.extensions',
 'sphinx_needs_enterprise.scripts',
 'sphinx_needs_enterprise.services']

package_data = \
{'': ['*'], 'sphinx_needs_enterprise': ['templates/*']}

install_requires = \
['azure-devops>=6.0.0-beta.4',
 'click>=7.0',
 'elasticsearch>=7.15',
 'jinja2>=2',
 'jira2markdown>=0.1',
 'licensing>=0.31',
 'm2r>=0.2',
 'requests>=2',
 'sphinxcontrib-needs>=0.7.2',
 'sphinxcontrib-programoutput>=0.17',
 'tqdm>=4.40']

entry_points = \
{'console_scripts': ['sne = sphinx_needs_enterprise.scripts.cli:cli']}

setup_kwargs = {
    'name': 'sphinx-needs-enterprise',
    'version': '0.1.3',
    'description': '',
    'long_description': 'Sphinx-Needs Enterprise\n=======================\n\n.. image:: docs/_static/sphinx-needs-enterprise-logo.png\n   :align: center\n   :width: 41%\n\n``Sphinx-Needs Enterprise`` is a collection of additional functions for ``Sphinx-Needs``, which are useful mostly\nin company-based environments.\n\nThis repository contains the complete source code and documentation.\n\nLicense\n-------\nThis package is released under the Business Source License, which allows free usage for private projects and sets\nsome duties for commercial projects. For details please take a look into our\n`LICENSE <https://raw.githubusercontent.com/useblocks/sphinx-needs-enterprise/main/LICENSE>`_ file.\n\n|\n| **To use Sphinx-Needs-Enterprise in commercial projects a commercial license is needed.**\n| Please get it contact with `useblocks <https://useblocks.com>`_ or visit sphinx-needs.com.\n',
    'author': 'team useblocks',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://github.com/useblocks/sphinx-needs-enterprise',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from jira2markdown import convert as jira_convert
from m2r import convert as md_convert

from sphinx_needs_enterprise.extensions.extension import ServiceExtension
from sphinx_needs_enterprise.util import dict_undefined_set


class JiraService(ServiceExtension):
    options = ["query", "prefix"]

    def __init__(self, app, name, config, **kwargs):
        self.app = app
        self.name = name

        # Set default values, if nothing got configured
        dict_undefined_set(config, "url", "http://127.0.0.1:8080")
        dict_undefined_set(config, "query", "")
        dict_undefined_set(config, "id_prefix", "JIRA_")
        dict_undefined_set(config, "url_postfix", "/rest/api/2/search")
        dict_undefined_set(config, "content", "{{data.fields.description}}")
        dict_undefined_set(config, "convert_content", True)

        # If no mapping is given, we need to use a need-type, which really exists.
        # So we take the first one from config, instead of hard-coding a value or
        # getting it from the jira issue.
        default_need_type = str(self.app.config["needs_types"][0]["directive"])

        mappings_default = {
            "id": ["key"],
            "type": default_need_type,
            "title": ["fields", "summary"],
            "status": ["fields", "status", "name"],
        }
        dict_undefined_set(config, "mappings", mappings_default)

        mappings_replaces_default = {}
        dict_undefined_set(config, "mappings_replaces", mappings_replaces_default)

        super().__init__(config, **kwargs)

    def request(self, options=None):
        params = self._prepare_request(options)

        request_params = {
            "method": "GET",
            "url": params["url"],
            "auth": params["auth"],
            "params": {"jql": params["query"]},
        }

        answer = self._send_request(request_params)
        data = answer.json()["issues"]

        if self.config["convert_content"]:
            # We need to transform the description text format to rst, before we proceed to extract
            # the final data
            for datum in data:
                if datum["fields"]["description"] is None:
                    datum["fields"]["description"] = ""
                    continue
                try:
                    # Convert from Jira format to markdown and from markdown to rst.
                    datum["fields"]["description"] = md_convert(jira_convert(datum["fields"]["description"]))
                except KeyError:
                    pass

        need_data = self._extract_data(data, options)

        return need_data


class InvalidConfigException(BaseException):
    pass
